/* lpr.c  26-Jan-89  A.J.Travis */

/*
 * print files, with tabs expanded into spaces
 */

#include <stdio.h>

#define TRUE 1
#define FALSE 0
#define SAME 0

#if MSDOS
#define DFS 0
#define EOL '\n'			/* end of line */
#else
#define DFS 1
#define ENABLE 2			/* enable printer */
#define DISABLE 3			/* disable printer */
#define EOL '\r'			/* end of line */
#endif

#define TABSIZ 8			/* tab width */
#define LINESIZ 81			/* line length + 1 */
#define LINEMAX 80			/* line length */
#define PAGESIZ 56			/* printed lines per page */
#define HALFPAGE ((PAGESIZ + 10) / 2)	/* lines per half-page */
#define PRINTER stdout			/* printer channel */

FILE *in;				/* input file */
char buf[LINESIZ];			/* i/o buffer */
char *bp;				/* buffer pointer */
int c;					/* ASCII character */
int col;				/* print column */
int line;				/* print line */
int page;				/* print page */
int paginate;				/* paginate output */

int main(argc, argv)
int argc;
char *argv[];
{
	if (argc < 2)
		usage();
	paginate = FALSE;		/* default is no pagination */
	while ((*argv[1] & 0xFF) == '-') {
		if (strcmp(argv[1], "-p") == SAME) {
			paginate++;
			--argc;
			argv++;
		}
		else
			usage();
	}
	if (argc < 2)
		usage();
	if ((in = fopen(argv[1], "r")) == NULL) {
		fprintf(stderr, "lpr: can't open %s\n", argv[1]);
		exit(-1);
	}
	col = 1;
	line = 1;
	page = 1;
	bp = buf;
#if DFS
	putc(ENABLE, PRINTER);
#endif
	if (paginate == FALSE)
		putc('\f', PRINTER);
	while ((c = getc(in)) != EOF) {
		if (c == '\t') {
			do {
				*bp++ = ' ';
			} while (++col % TABSIZ != 1 & col < LINESIZ);
		}
		else if (c == EOL | col == LINEMAX) {
			if (col == LINEMAX)
				*bp++ = c;
			if (line++ % PAGESIZ == 1 & paginate == TRUE) {
				fprintf(PRINTER, "\f\n\n%-70s", argv[1]);
				fprintf(PRINTER, "page %d\n\n\n", page++);
			}
			*bp = '\0';
			fprintf(PRINTER, "%s\n", buf);
			col = 1;
			bp = buf;
		}
		else {
			*bp++ = c;
			col++;
 		}
	}
	fprintf(PRINTER, "\f***** page alignment *****");
	for (c = 0; c < HALFPAGE; c++)
		putc('\n', PRINTER);
#if DFS
	putc(DISABLE, PRINTER);
#endif
	close(in);
	exit(0);
}

/*
 * report correct usage and exit
 */
usage()
{
	fprintf(stderr, "usage: lpr [-p] file\n");
	exit(-1);
}
