/* tccom.c  -  'Tiny' C compiler for the 6502 */

/* 11-Apr-1989  v0.70 - A.J.Travis */
/* 22-Jun-1991  v0.71 - J.G.Harston
 *   main() ends with exit(0);
 */

#include <stdio.h>

/*
 * Boolean flags
 */
#define TRUE 1
#define FALSE 0
#define ERROR -1
#define UNDEFINED NULL
#define SAME 0

/*
 * Implementation dependent definitions
 */
#define INTWIDTH 2	/* integer size */
#define CHARWIDTH 1	/* char size */

/*
 * symbol length
 */
#define NAMESIZE 8
#define NAMEMAX 7 	/* NAMESIZE - 1 */

/*
 * symbol table entry format
 */
#define NAME		0		/* null terminated string */
#define IDENT		8		/* NAMESIZE */
#define TYPE		9		/* IDENT + 1 */
#define OFFSET		10		/* TYPE + 1 */

/*
 * symbol table parameters
 */
#define SYMSIZ		12		/* NAMESIZE + 4 */
#define NUMGLBS		180
#define NUMLOCS		20
#define SYMTBSZ		2400	/* (NUMGLBS + NUMLOCS) * SYMSIZ */
#define STARTGLB	symtab
#define ENDGLB		symtab + 2160	/* NUMGLBS * SYMSIZ */
#define STARTLOC	symtab + 2172	/* (NUMGLBS * SYMSIZ) + SYMSIZ */
#define ENDLOC		symtab + 2388	/* SYMTBSZ - SYMSIZ */

/*
 * symbol "IDENT"
 */
#define VARIABLE 1
#define ARRAY 2
#define POINTER 3
#define FUNCTION 4

/*
 * symbol "TYPE"
 */
#define CCHAR 1
#define CINT 2

/*
 * L-value definition
 */
#define LSIZE 2		/* size of L-value structure */
#define SYMBOL 0	/* lval[0] symbol table address (0 for constant) */
#define CTYPE 1		/* lval[1] type of indirect object (0 for static) */

/*
 * "while" statement queue
 */
#define WQTABSZ 16
#define WQSIZ 4
#define WQMAX wq + 12	/* WQTABSZ - WQSIZ */

/*
 * entry offsets in while queue
 */
#define WQSYM 0
#define WQSP 1
#define WQLOOP 2
#define WQLAB 3

/*
 * input line
 */
#define LINESIZE 80
#define LINEMAX LINESIZE - 1

/*
 * statement types (tokens)
 */
#define STIF 1
#define STWHILE 2
#define STRETURN 3
#define STBREAK 4
#define STCONT 5
#define STGOTO 6
#define STEXP 7
#define STEXIT 8

/*
 * expression type
 */
#define ARITH 0
#define REL 1

/*
 * error codes
 */
#define E_DECL		 1		/* Illegal function or declaration */
#define E_DEF		 2		/* Already defined */
#define E_LPAREN	 3		/* Missing opening parenthesis */
#define E_ARGNAME	 4		/* Illegal argument name */
#define E_RPAREN	 5		/* Missing closing parenthesis */
#define E_ARGNUM	 6		/* Wrong number of arguments */
#define E_NAME1	 7		/* Illegal symbol name */
#define E_ARG		 8		/* Argument name expected */
#define E_COMMA	 9		/* Comma expected */
#define E_WHILE	10		/* 'while' expected */
#define E_LABEL	11		/* Illegal 'goto' label */
#define E_LVAL		12		/* L-value required */
#define E_ADDRESS	13		/* Illegal address */
#define E_CONSTANT	14		/* Can't subscript constant */
#define E_VARIABLE	15		/* Can't subscript variable */
#define E_EXPR		16		/* Illegal expression */
#define E_DELIM	17		/* delimeter expected */
#define E_NAME2	18		/* Illegal symbol name */
#define E_GFULL	19		/* Global symbol table full */
#define E_LFULL	20		/* Local symbol table full */
#define E_XWHILE	21		/* Too many active whiles */
#define E_NOWHILE	22		/* No active whiles */
#define E_CCONST	23		/* Illegal character constant */
#define E_NEEDCONST	24		/* Constant required */
#define E_SIZE		25		/* Negative size illegal */
#define E_SEMI		26		/* Missing semicolon */
#define E_BRACKET	27		/* Missing bracket */

/*
 * function types
 */
char *addglb();
char *findglb();
char *addloc();
char *findloc();
int *readwhile();

/*
 * global variables
 */
char symtab[SYMTBSZ];		/* symbol table	*/
char *glbp;			/* global symbol pointer */
char *locp;			/* local symbol pointer */

int wq[WQTABSZ];		/* while queue */
int *wqp;			/* while queue pointer */

int litp;			/* literal pointer */

char ibuf[LINESIZE];		/* input buffer */
char *ip;
char *ipmax;
char c1, c2;			/* input characters */

int nxtlab;			/* next available label # */
int litlab;			/* label assigned to literal pool */
int sp;			/* stack pointer */
int csp;			/* current stack pointer */
int argstk;			/* function argument sp	*/
int ncmp;			/* # open compound statements */
int eof;			/* non-zero on final input eof */
FILE *lf;			/* fp for literal file */
FILE *in;			/* fp for input file	*/
FILE *out;			/* fp for output file */
int lastst;			/* last executed statement type */
int lastop;			/* last executed operator type */
int hasmain;			/* does this file have main() */
int ctext;			/* show C source as assembler comments */
char *cp;			/* work ptr to any char buffer */

int main(argc, argv)
int argc;
char *argv[];
{
	ctext = FALSE;
	while ((*argv[1] & 0xFF) == '-') {	/* BUG in compiler */
		if (strcmp(argv[1], "-C") == SAME)
			ctext = TRUE;
		else
			usage();
		--argc;
		argv++;
	}
	in = out = UNDEFINED;
	if ((argc < 2) | (argc > 3))
		usage();
	if ((lf = fopen("litfile", "wb")) == NULL)
		cant("litfile");
	if ((in = fopen(argv[1], "r")) == NULL)
		cant(argv[1]);
	if (argc == 2)
		out = stdout;
	else {
		if ((out = fopen(argv[2], "w")) == NULL)
			cant(argv[2]);
	}
	glbp = STARTGLB;
	locp = STARTLOC;
	wqp = wq;
	sp = eof = ncmp = lastst = litp = 0;
	nxtlab = 1;
	litlab = getlabel();
	reset();
	header();
	parse();
	dumplits();
	dumpglbs();
	trailer();
	fclose(lf);
	fclose(in);
	if (out != stdout)
		fclose(out);
	exit(0);
}

/*
 * report correct usage, and exit
 */
usage()
{
	fprintf(stderr, "usage: tccom [-C] infile [outfile]\n");
	exit(-1);
}

/*
 * Process all input text
 * ----------------------
 * At this level, only static declarations and function definitions are legal
 */
parse()
{
	while (eof == FALSE) {
		if (token("char"))
			declglb(CCHAR);
		else if (token("int"))
			declglb(CINT);
		else if (token("extern"))
			/* ignore extern */
			skipto(';');
		else
			/* assume function returning int */
			declglb(CINT);
		skip();
	}
}

/*
 * New function definition
 */
newfunc(name)
char *name;				/* function name */
{
	int argtop;
	char n[NAMESIZE];

	locp = STARTLOC;
	argstk = 0;
	while (match(')') == FALSE) {
		if (symname(n) == FALSE)
			error(E_ARGNAME);
		else {
			if (findloc(n))
				error(E_DEF);
			else {
				addloc(n, 0, 0, argstk);
				argstk = argstk + INTWIDTH;
			}
		}
		skip();
		if (c1 != ')') {
			if (match(',') == FALSE)
				error(E_RPAREN);
		}
		if (needstend())
			break;
	}

	/* return NULL if this is function declaration */
	skip();
	if (c1 == ';')
		return NULL;

	/* print name of each function as it is processed */
	printf("%s\n", name);

	hasmain = (strcmp(name, "main") == SAME);
	label(name);
#ifdef DEBUG
	trace();
#endif
	sp = 0;
	argtop = argstk;
	while (argstk) {
		if (token("char")) {
			getarg(CCHAR, argtop);
			needsemi();
		}
		else if (token("int")) {
			getarg(CINT, argtop);
			needsemi();
		}
		else
			error(E_ARGNUM);
	}
	statement();
	if (hasmain == FALSE & lastst != STRETURN) {
		modstk(0);
		ret();
	}
	else if (hasmain == TRUE & lastst != STEXIT) {
		modstk(0);
		c_exit();
	}
	sp = 0;
	locp = STARTLOC;
	return FUNCTION;
}

/*
 * Get function arguments
 */
getarg(t, argtop)
int t, argtop;
{
	int j, address;
	char *argp, n[NAMESIZE];

	while(TRUE) {
		if (argstk == 0)
			return;
		if (match('*'))
			j = POINTER;
		else
			j = VARIABLE;
		if (symname(n) == FALSE)
			error(E_NAME1);
		if (match('[')) {
			while (inbyte() != ']')
				if (needstend())
					break;
			if ((t == CCHAR) & (j == POINTER))
				t = CINT;
			j = POINTER;
		}
		if ((argp = findloc(n)) == 0)
			error(E_ARG);
		argp[IDENT] = j;
		argp[TYPE] = t;
		argstk = argstk - INTWIDTH;
		if (needstend())
			return;
		if (match(',') == FALSE)
			error(E_COMMA);
	}
}

/*
 * Process a statement
 */
statement()
{
	lastst = UNDEFINED;
	csp = sp;
	while (TRUE) {
		if (token("char")) {
			declloc(CCHAR);
			needsemi();
		}
		else if (token("int")) {
			declloc(CINT);
			needsemi();
		}
		else
			break;
	}
	sp = modstk(csp);
	if (match('{'))
		compound();
	else if (token("if"))
		doif();
	else if (token("while"))
		dowhile();
	else if (token("for"))
		dofor();
	else if (token("do")) {
		dodo();
		needsemi();
	}
	else if (token("goto")) {
		dogoto();
		needsemi();
	}
	else if (dolab())
		return;
	else if (token("return")) {
		doreturn();
		needsemi();
		lastst = STRETURN;
	}
	else if (token("break")) {
		dobreak();
		needsemi();
	}
	else if (token("continue")) {
		docont();
		needsemi();
	}
	else if (token("exit")) {
		doexit();
		needsemi();
		lastst = STEXIT;
	}
	else {
		expression();
		needsemi();
	}
}

/*
 * Process compound statement
 */
compound()
{
	++ncmp;
	while (match('}') == FALSE) {
		if (eof)
			return;
		statement();
	}
	--ncmp;
}

/*
 * Process IF statement
 */
doif()
{
	char *flev;
	int fsp, flab1, flab2;

	flev = locp;
	fsp = sp;
	flab1 = getlabel();
	test(flab1);
	statement();
	sp = modstk(fsp);
	locp = flev;
	if (token("else") == FALSE) {
		outlab(flab1);
		return;
	}
	if (lastst == STRETURN) {
		outlab(flab1);
		statement();
		sp = modstk(fsp);
		locp = flev;
		return;
	}
	jump(flab2 = getlabel());
	outlab(flab1);
	statement();
	sp = modstk(fsp);
	locp = flev;
	outlab(flab2);
}

/*
 * Process WHILE statement
 */
dowhile()
{
	int wq[4];

	*wq = locp;
	wq[WQSP] = sp;
	wq[WQLOOP] = getlabel();
	wq[WQLAB] = getlabel();
	addwhile(wq);
	outlab(wq[WQLOOP]);
	test(wq[WQLAB]);
	statement();
	jump(wq[WQLOOP]);
	outlab(wq[WQLAB]);
	locp = *wq;
	sp = modstk(wq[WQSP]);
	delwhile();
}

dofor()
{
	int wq[4];
	char control[LINESIZE];		/* loop control string */
	char trail[LINESIZE];		/* trailing text after ')' */

	*wq = locp;
	wq[WQSP] = sp;
	wq[WQLOOP] = getlabel();
	wq[WQLAB] = getlabel();
	addwhile(wq);
	needbrack('(');
	expression();
	needsemi();
	outlab(wq[WQLOOP]);
	fortest(wq[WQLAB]);
	needsemi();
	strcpy(control, ip);
	skipto(')');
	statement();
	strcpy(trail, ip);
	strcpy(ibuf, control);
	ip = ibuf;
	expression();
	needbrack(')');
	jump(wq[WQLOOP]);
	outlab(wq[WQLAB]);
	locp = *wq;
	sp = modstk(wq[WQSP]);
	delwhile();
	strcpy(ibuf, trail);
	ip = ibuf;
}

dodo()
{
	int wq[4];

	*wq = locp;
	wq[WQSP] = sp;
	wq[WQLOOP] = getlabel();
	wq[WQLAB] = getlabel();
	addwhile(wq);
	outlab(wq[WQLOOP]);
	statement();
	if (token("while") == FALSE)
		error(E_WHILE);
	dotest(wq[WQLOOP]);
	outlab(wq[WQLAB]);
	locp = *wq;
	sp = modstk(wq[WQSP]);
	delwhile();
}

dogoto()
{
	char s[LINESIZE];

	if (symname(s) == FALSE)
		error(E_LABEL);
	sjump(s);
}

/*
 * Process RETURN statement
 */
doreturn()
{
	if (needstend() == FALSE)
		expression();
	modstk(0);
	ret();
}

/*
 * Process BREAK statement
 */
dobreak()
{
	int *p;

	if ((p = readwhile()) == 0)
		return;
	modstk(p[WQSP]);
	jump(p[WQLAB]);
}

/*
 * Process CONTINUE statement
 */
docont()
{
	int *p;

	if ((p = readwhile()) == 0)
		return;
	modstk(p[WQSP]);
	jump(p[WQLOOP]);
}

dolab()
{
	char *savech;			/* saved position of ip */
	char n[LINESIZE];		/* label name */

	savech = ip;
	if (symname(n) == FALSE)
		return FALSE;
	if (match(':') == FALSE) {
		ip = savech;
		return FALSE;
	}
	label(n);
	return TRUE;
}

doexit()
{
	if (needstend() == FALSE)
		expression();
	modstk(0);
	c_exit();
}

/*
 * Expression evaluation by recursive descent
 */
expression()
{
	int lval[LSIZE];

	lastop = 0;
	if (c1 == ';') {
		if (ctext)
			comment("null expression\n");
	}
	else if (hier1(lval))
		rvalue(lval);
}

hier1(lval)
int *lval;
{
	int k;
	int lval2[LSIZE];

	k = hier2(lval);
	if (match('=')) {
		if (k == 0)
			error(E_LVAL);
		if (lval[CTYPE])
			push();
		if (hier1(lval2))
			rvalue(lval2);
		store(lval);
		return 0;
	}
	else
		return k;
}

hier2(lval)
int *lval;
{
	int k;
	int lval2[LSIZE];

	k = hier3(lval);
	skip();
	if (c1 != '|')
		return k;
	if (k)
		rvalue(lval);
	while(TRUE) {
		if (match('|')) {
			push();
			if (hier3(lval2))
				rvalue(lval2);
			or();
		}
		else
			return 0;
	}
}

hier3(lval)
int *lval;
{
	int k;
	int lval2[LSIZE];

	k = hier4(lval);
	skip();
	if (c1 != '^')
		return k;
	if (k)
		rvalue(lval);
	while(TRUE) {
		if (match('^')) {
			push();
			if (hier4(lval2))
				rvalue(lval2);
			xor();
		}
		else
			return 0;
	}
}

hier4(lval)
int *lval;
{
	int k;
	int lval2[LSIZE];

	k = hier5(lval);
	skip();
	if (c1 != '&')
		return k;
	if (k)
		rvalue(lval);
	while(TRUE) {
		if (match('&')) {
			push();
			if (hier5(lval2))
				rvalue(lval2);
			and();
		}
		else
			return 0;
	}
}

/*
 * check for == and !=
 */
hier5(lval)
int *lval;
{
	int k;
	int lval2[LSIZE];

	k = hier6(lval);
	skip();
	if (c2 != '=')
		return k;
	if ((c1 != '=') & (c1 != '!'))
		return k;
	if (k)
		rvalue(lval);
	while(TRUE) {
		if (amatch("==")) {
			push();
			if (hier6(lval2))
				rvalue(lval2);
			eq();
		}
		else if (amatch("!=")) {
			push();
			if (hier6(lval2))
				rvalue(lval2);
			ne();
		}
		else
			return 0;
	}
}


/*
 * wish to identify >, <, >=, <=, but reject >>, <<
 */
hier6(lval)
int *lval;
{
	int k;
	int lval2[LSIZE];

	k = hier7(lval);
	skip();
	if ((c1 != '<') & (c1 != '>'))
		return k;
	if ((c2 == '<') | (c2 == '>'))
		return k;
	if (k)
		rvalue(lval);
	while(TRUE) {
		if (amatch("<=")) {
			push();
			if (hier7(lval2))
				rvalue(lval2);
			le(lval, lval2);
		}
		else if (amatch(">=")) {
			push();
			if (hier7(lval2))
				rvalue(lval2);
			ge(lval, lval2);
		}
		else if ((c1 == '<') & (c2 != '<')) {
			inbyte();
			push();
			if (hier7(lval2))
				rvalue(lval2);
			lt(lval, lval2);
		}
		else if ((c1 == '>') & (c2 != '>')) {
			inbyte();
			push();
			if (hier7(lval2))
				rvalue(lval2);
			gt(lval, lval2);
		}
		else
			return 0;
	}
}

hier7(lval)
int *lval;
{
	int k;
	int lval2[LSIZE];

	k = hier8(lval);
	skip();
	if ((c1 != '>') & (c1 != '<'))
		return k;
	if ((c2 != '>') & (c2 != '<'))
		return k;
	if (k)
		rvalue(lval);
	while(TRUE) {
		if (amatch(">>")) {
			push();
			if (hier8(lval2))
				rvalue(lval2);
			asr();
		}
		else if (amatch("<<")) {
			push();
			if (hier8(lval2))
				rvalue(lval2);
			asl();
		}
		else
			return 0;
	}
}

hier8(lval)
int *lval;
{
	int k;
	int lval2[LSIZE];

	k = hier9(lval);
	skip();
	if ((c1 != '+') & (c1 != '-'))
		return k;
	if (k)
		rvalue(lval);
	while(TRUE) {
		if (match('+')) {
			push();
			if (hier9(lval2))
				rvalue(lval2);
			add(lval);
		}
		else if (match('-')) {
			push();
			if (hier9(lval2))
				rvalue(lval2);
			sub(lval);
		}
		else
			return 0;
	}
}

hier9(lval)
int *lval;
{
	int k;
	int lval2[LSIZE];

	k = hier10(lval);
	skip();
	if ((c1 != '*') & (c1 != '/') & (c1 != '%'))
		return k;
	if (k)
		rvalue(lval);
	while(TRUE) {
		if (match('*')) {
			push();
			if (hier9(lval2))
				rvalue(lval2);
			mult();
		}
		else if (match('/')) {
			push();
			if (hier10(lval2))
				rvalue(lval2);
			div(lval, lval2);
		}
		else if (match('%')) {
			push();
			if (hier10(lval2))
				rvalue(lval2);
			mod(lval, lval2);
		}
		else
			return 0;
	}
}

hier10(lval)
int *lval;
{
	int k;
	char *p;

	if (amatch("++")) {
		if ((k = hier10(lval)) == 0)
			error(E_LVAL);
		if (lval[CTYPE])
			push();
		rvalue(lval);
		inc(lval);
		store(lval);
		return 0;
	}
	else if (amatch("--")) {
		if ((k = hier10(lval)) == 0)
			error(E_LVAL);
		if (lval[CTYPE])
			push();
		rvalue(lval);
		dec(lval);
		store(lval);
		return 0;
	}
	else if (match('-')) {
		if ((k = hier10(lval)) != NULL)
			rvalue(lval);
		neg();
		return 0;
	}
	else if (match('*')) {
		if ((k = hier10(lval)) != NULL)
			rvalue(lval);
		lval[CTYPE] = CINT;
		if ((p = lval[SYMBOL]) != NULL)
			lval[CTYPE] = p[TYPE];
		lval[SYMBOL] = 0;
		return 1;
	}
	else if (match('&')) {
		if ((k = hier10(lval)) == 0)
			error(E_ADDRESS);
		else if (lval[CTYPE])
			return 0;
		else {
			immediate(p = lval[SYMBOL]);
			lval[CTYPE] = p[TYPE];
			return 0;
		}
	}
	else {
		k = hier11(lval);
		if (amatch("++")) {
			if (k == 0)
				error(E_LVAL);
			if (lval[CTYPE])
				push();
			rvalue(lval);
			inc(lval);
			store(lval);
			dec(lval);
			return 0;
		}
		else if (amatch("--")) {
			if (k == 0)
				error(E_LVAL);
			if (lval[CTYPE])
				push();
			rvalue(lval);
			dec(lval);
			store(lval);
			inc(lval);
			return 0;
		}
		else
			return k;
	}
}

hier11(lval)
int *lval;
{
	int k;
	char *p;

	k = primary(lval);
	p = lval[SYMBOL];
	skip();
	if ((c1 == '[') | (c1 == '('))
		while (TRUE) {
			if (match('[')) {
				if (p == NULL)
					error(E_CONSTANT);
				else if (p[IDENT] == POINTER)
					rvalue(lval);
				else if (p[IDENT] != ARRAY)
					error(E_VARIABLE);
				push();
				expression();
				needbrack(']');
				add(lval);
				lval[SYMBOL] = NULL;
				lval[CTYPE] = p[TYPE];
				k = 1;
			}
			else if (match('(')) {
				if (p == NULL)
					callfunction(NULL);
				else if (p[IDENT] == FUNCTION)
					callfunction(p);
				else {
					rvalue(lval);
					callfunction(NULL);
				}
				lval[SYMBOL] = NULL;
				k = 0;
			}
			else
				return k;
		}
	if (p == NULL)
		return k;
	if (p[IDENT] == FUNCTION) {
		immediate(p);
		return 0;
	}
	return k;
}

primary(lval)
int *lval;
{
	int k;
	int num[1];
	char *p, sname[NAMESIZE];

	/* check for parenthesised expression */
	if (match('(')) {
		k = hier1(lval);
		needbrack(')');
		return k;
	}

	if (symname(sname)) {
		if ((p = findloc(sname)) != NULL) {
			getloc(p);
			lval[SYMBOL] = p;
			lval[CTYPE] = p[TYPE];
			if (p[IDENT] == POINTER)
				lval[CTYPE] = CINT;
			if (p[IDENT] == ARRAY)
				return 0;
			else
				return 1;
		}
		if ((p = findglb(sname)) != NULL)
			if (p[IDENT] != FUNCTION) {
				lval[SYMBOL] = p;
				lval[CTYPE] = 0;
				if (p[IDENT] != ARRAY)
					return 1;
				immediate(p);
				lval[CTYPE] = p[TYPE];
				return 0;
			}
		p = addglb(sname, FUNCTION, CINT, 0);
		lval[SYMBOL] = p;
		lval[CTYPE] = 0;
		return 0;
	}
	if (constant(num))
		return lval[SYMBOL] = lval[CTYPE] = 0;
	else
		error(E_EXPR);
}

/*
 * Process function call
 */
callfunction(name)
char *name;
{
	int nas;
	char *p1, *p2;

	skip();
	p1 = ip;
	while (c1 != ')') {
		if (c1 == '"')
			skipto('"');
		else if (c1 == '\'')
			skipto('\'');
		else if (c1 == '(')
			parscan();
		else if (c1 == '\0')
			break;	
		else
			bump();
	}
	needbrack(')');
	p2 = ip;
	ip = p1;
	skip();
	if (name == NULL) {
		push();
	}
	nas = arglist(name, 0);
	if (name)
		call(name);
	else
		callstk();
	sp = modstk(sp + nas);
	ip = p2;
	skip();
}

/*
 * recursively extract argument list
 */
arglist(n, nas)
char *n;		/* name of calling function (NULL if indirect call) */
int nas;		/* number of arguments */
{
	char *p;

	p = ip;
	if (argq())
		nas = arglist(n, nas);
	ip = p;
	skip();
	if (c1 == ')')
		return nas;
	expression();
	if (n == NULL)
		swapstk();
	push();
	return nas + INTWIDTH;
}

/*
 * extract one argument from queue
 */
argq()
{
	int c;

	skip();
	while ((c1 != ',') & (c1 != ')')) {
		if (c1 == '"')
			skipto('"');
		else if (c1 == '\'')
			skipto('\'');
		else if (c1 == '(')
			parscan();
		else
			bump();
	}
	c = c1;
	bump();
	return c == ',';
}

skipto(delim)
char delim;
{
	while ((bump()) != delim) {
		if (c1 == '\\')
			bump();
		if (c1 == '\0')
			error(E_DELIM);
	}
	bump();
}

parscan()
{
	int nest;

	nest = 1;
	do {
		if ((bump()) == '(')
			++nest;
		else if (c1 == ')')
			--nest;
	}
	while ((nest > 0) & (c1 != ')'));
	bump();
}

/*
 * Declare a static variable
 * -------------------------
 * makes an entry in the symbol table so that subsequent
 * reference can call symbol by name
 */
declglb(typ)
int typ;
{
	int j, k;
	char *p;
	char sname[NAMESIZE];

	do {
		if (needstend())
			return;
		k = 1;
		if (match('*'))
			j = POINTER;
		else
			j = VARIABLE;
		if (symname(sname) == FALSE)
			error(E_NAME2);
		if ((p = findglb(sname)) != NULL) {
			if ((p[IDENT] != FUNCTION) | (p[OFFSET] == FUNCTION))
				error(E_DEF);
		}
		if (match('[')) {
			if ((k = needsub()) == 0)
				j = POINTER;
			else
				j = ARRAY;
		}
		else if (match('(')) {
			k = newfunc(sname);
			j = FUNCTION;
		}
		addglb(sname, j, typ, k);
		if ((j == FUNCTION) & (k == FUNCTION))
			return;
	}
	while (match(','));
	needsemi();
}

/*
 * Declare local variables
 */
declloc(typ)
int typ;
{
	int k, j;
	char sname[NAMESIZE];

	do {
		if (needstend())
			return;
		if (match('*'))
			j = POINTER;
		else
			j = VARIABLE;
		if (symname(sname) == FALSE)
			error(E_NAME2);
		if (findloc(sname))
			error(E_DEF);
		if (match('[')) {
			if ((k = needsub()) == 0) {
				j = POINTER;
				k = INTWIDTH;
			}
			else {
				j = ARRAY;
				if (typ == CINT)
					k = k * INTWIDTH;
			}
		}
		else if ((typ == CCHAR) & (j != POINTER))
			k = CHARWIDTH;
		else
			k = INTWIDTH;
		csp = csp - k;
		addloc(sname, j, typ, csp);
	}
	while (match(','));
}

/*
 * Insert new global symbol name
 */
char *addglb(sname, id, typ, value)
char *sname, id, typ;
int value;
{
	char *p;

	if ((cp = findglb(sname)) != NULL)
		return cp;
	if (glbp >= ENDGLB)
		error(E_GFULL);
	cp = p = glbp;
	while ((*p++ = *sname++) != '\0')
		;
	cp[IDENT] = id;
	cp[TYPE] = typ;
	mputw(cp + OFFSET, value);
	glbp = glbp + SYMSIZ;
	return cp;
}

/*
 * Find a global symbol name
 */
char *findglb(sname)
char *sname;
{
	char *p;

	p = STARTGLB;
	while (p != glbp) {
		if (strcmp(sname, p) == SAME)
			return p;
		p = p + SYMSIZ;
	}
	return NULL;
}

/*
 * Insert new local symbol name
 */
char *addloc(sname, id, typ, value)
char *sname, id, typ;
int value;
{
	char *p;

	if ((cp = findloc(sname)) != NULL)
		return cp;
	if (locp >= ENDLOC)
		error(E_LFULL);
	cp = p = locp;
	while ((*p++ = *sname++) != '\0')
		;
	cp[IDENT] = id;
	cp[TYPE] = typ;
	mputw(cp + OFFSET, value);
	locp = locp + SYMSIZ;
	return cp;
}

/*
 * Find a local symbol name
 */
char *findloc(sname)
char *sname;
{
	char *p;

	p = STARTLOC;
	while (p != locp) {
		if (strcmp(sname, p) == SAME)
			return p;
		p = p + SYMSIZ;
	}
	return NULL;
}

/*
 * WHILE table manipulation
 */
addwhile(p)
int *p;
{
	int k;

	if (wqp > WQMAX)
		error(E_XWHILE);
	k = 0;
	while (k < WQSIZ)
		*wqp++ = p[k++];
}

delwhile()
{
	if (readwhile())
		wqp = wqp - WQSIZ;
}

int *readwhile()
{
	if (wqp == wq)
		error(E_NOWHILE);
	else
		return wqp - WQSIZ;
}

/*
 * Generate next label
 */
getlabel()
{
	return ++nxtlab;
}

/*
 * Read symbol name
 */
symname(sname)
char *sname;
{
	int k;

	skip();
	if ((isalpha(c1) == FALSE) & (c1 != '_'))
		return FALSE;
	k = 0;
	while (isalnum(c1) | c1 == '_') {
		if (k < NAMEMAX)
			sname[k++] = c1;
		bump();
	}
	sname[k] = '\0';
	return TRUE;
}

/*
 * Check for a number in input
 */
number(val)
int *val;
{
	int k, sign;

	k = sign = 1;
	while (k) {
		k = 0;
		if (match('+'))
			k = 1;
		if (match('-')) {
			sign = (-sign);
			k = 1;
		}
	}
	if (isdigit(c1) == FALSE)
		return getqchar(val);
	if (match('0')) {
		if (match('x')) {
			while (isxdigit(c1))
				k = k * 16 + toint(gch());
		}
		else {
			while (isodigit(c1))
				k = k * 8 + toint(gch());
		}
	}
	else {
		while (isdigit(c1))
			k = k * 10 + toint(gch());
	}
	if (sign < 0)
		k = (-k);
	*val = k;
	return TRUE;
}

/*
 * convert hex char to integer
 */
toint(c)
char c;
{
	if (isdigit(c))
		return c - '0';
	if ((c >= 'A') & (c <= 'F'))
		return 10 + c - 'A';
	if ((c >= 'a') & (c <= 'f'))
		return 10 + c - 'a';
	return -1;
}

/*
 * check for octal digits
 */
isodigit(c)
char c;
{
	return (c >= '0') & (c <= '8');
}

/*
 * Load a (positive) constant
 */
constant(val)
int *val;
{
	if (number(val))
		immed(*val);
	else if (getqstring(val))
		literal(litlab, *val);
	else
		return FALSE;
	return TRUE;
}

/*
 * Get one or two characters from input stream
 */
getqchar(val)
int *val;
{
	char c;

	if (match('\'') == FALSE)
		return FALSE;
	if (c1 == '\\')
		c = cesc();
	else
		c = gch();
	if (c1 != '\'')
		error(E_CCONST);
	gch();
	*val = c;
	return TRUE;
}

cesc()
{
	int c;

	gch();
	if (c1 == 'n')
		c = '\n';	/* newline */
	else if (c1 == 't')
		c = '\t';	/* tab */
	else if (c1 == 'b')
		c = '\b';	/* backspace */
	else if (c1 == 'r')
		c = '\r';	/* return */
	else if (c1 == 'f')
		c = '\f';	/* form feed */
	else if (c1 == '\\')
		c = '\\';	/* backslash */
	else if (isodigit(c1)) {
		c = toint(gch());
		while (isodigit(c1))
			c = c * 8 + toint(gch());
		return c;
	}
	else
		c = c1;	/* any other char */
	gch();
	return c;
}

/*
 * Get string from input stream
 */
getqstring(val)
int *val;
{
	char c;

	if (match('"') == FALSE)
		return FALSE;
	*val = litp;
	while (c1 != '"') {
		if (c1 == '\0')
			break;
		else if (c1 == '\\')
			c = cesc();
		else
			c = gch();
		putc(c, lf);
		++litp;
	}
	gch();
	putc('\0', lf);
	++litp;
	return TRUE;
}

/*
 * Compare literal with line buffer contents
 * -----------------------------------------
 * advances buffer pointer if literal matches
 */
match(lit)
char lit;
{
	skip();
	if (c1 != lit)
		return FALSE;
	else {
		bump();
		return TRUE;
	}
}

/*
 * match n chars with input buffer
 */
amatch(lit)
char *lit;
{
	int len;

	len = strlen(lit);
	skip();
	if (strncmp(ip, lit, len) != SAME)
		return FALSE;
	else {
		ip = ip + len;
		c1 = *ip;
		return TRUE;
	}
}

/*
 * match token with input buffer
 */
token(lit)
char *lit;
{
	int len;

	skip();
	len = strlen(lit);
	if (strncmp(ip, lit, len) != SAME)
		return FALSE;
	else if (isalnum(ip[len]))
		return FALSE;
	else {
		ip = ip + len;
		c1 = *ip;
		return TRUE;
	}
}

/*
 * Get array bounds
 */
needsub()
{
	int num[1];

	if (match(']'))
		return 0;
	if (number(num) == 0)
		error(E_NEEDCONST);
	if (*num < 0)
		error(E_SIZE);
	needbrack(']');
	return *num;
}

/*
 * Check for semicolon
 */
needsemi()
{
	if (match(';') == FALSE)
		error(E_SEMI);
}

/*
 * Check for end of statement
 */
needstend()
{
	skip();
	return (c1 == ';') | (c1 == '\0');
}

needbrack(c)
char c;
{
	skip();
	if (match(c) == FALSE)
		error(E_BRACKET);
}

/*
 * Skip white space in input
 */
skip()
{
	while (isspace(*ip) | *ip == '\0') {
		if (*ip++ == '\0')
			inline();
	}
	c1 = *ip;
	c2 = *(ip + 1);
}

/*
 * bump lexical input
 */
bump()
{
	c1 = *++ip;
	c2 = *(ip + 1);
	return c1;
}

/*
 * Get a character
 * ---------------
 * reads in a line if necessary
 */
inbyte()
{
	while(*ip == '\0') {
		if (eof)
			return '\0';
		inline();
	}
	return gch();
}

/*
 * Read in a line
 */
inline()
{
	if ((ip = fgets(ibuf, LINESIZE, in)) == NULL)
		eof = TRUE;
	else if (ctext)
		comment(ip);
}

/*
 * Reset input Buffer
 */
reset()
{
	ip = ibuf;
	c1 = c2 = *ip = '\0';
}

/*
 * Get next input character
 */
gch()
{
	int c;

	if (*ip == '\0')
		return '\0';
	else {
		c = *ip;
		bump();
		return c;
	}
}

/*
 * put 16-bit integer into memory word
 */
mputw(p, val)
int *p;				/* (int *) cast of char *p */
int val;
{
	*p = val;
}

/*
 * get 16-bit integer word from memory
 */
mgetw(p)
int *p;				/* (int *) cast of char *p */
{
	return *p;
}

/*
 * Report type and position of an error in the source line
 */
error(n)
int n;
{
	char *p;

#if MSDOS
	fflush(stdout);
#endif
	fprintf(stderr, "\n%s", ibuf);
	p = ibuf;
	while (p < ip) {
		if (*p == '\t')
			fprintf(stderr, "\t");
		else
			fprintf(stderr, " ");
		++p;
	}
	fprintf(stderr, "^\nerror: %d\n", n);
	exit(-1);
}

/*
 * can't open ...
 */
cant(s)
char *s;	/* filename */
{
	fprintf(stderr, "tccom: can't open %s\n", s);
	exit(-1);
}

/*
 * Evaluate condition
 */
test(label)
int label;
{
	needbrack('(');
	expression();
	needbrack(')');
	if (lastop == REL)
		testjump(label);
	else
		testnz(label);
}

fortest(label)
int label;
{
	expression();
	if (lastop == REL)
		testjump(label);
	else
		testnz(label);
}

dotest(label)
int label;
{
	needbrack('(');
	expression();
	needbrack(')');
	if (lastop == REL)
		dotestjump(label);
	else
		dotestnz(label);
}

/*
 * Store a value in memory
 */
store(lval)
int *lval;
{
	if (lval[CTYPE] == 0)
		putmem(lval);
	else {
		pop();
		putstk(lval[CTYPE]);
	}
}

/*
 * Get a value from memory
 */
rvalue(lval)
int *lval;
{
	if ((lval[SYMBOL] != NULL) & (lval[CTYPE] == 0))
		getmem(lval);
	else
		indirect(lval[CTYPE]);
}

/*
 * Dump the literal pool
 * ---------------------
 * if nothing there, exit
 */
dumplits()
{
	int k, c;

	if (litp) {
		outlab(litlab);
		fclose(lf);
		if ((lf = fopen("litfile", "rb")) == NULL)
			cant("litfile");
		for (k = 0; k < litp; ++k)
			defbyte(getc(lf));
	}
	fprintf(out, "~eot\n");
}

/*
 * Dump all static variables
 */
dumpglbs()
{
	int sz;
	int dp;

	dp = 0;
	fprintf(out, "~data =~eot\n");
	cp = STARTGLB;
	while (cp < glbp) {
		if (cp[IDENT] != FUNCTION) {
			fprintf(out, "_%s", cp);
			sz = mgetw(cp + OFFSET);
			if ((cp[TYPE] == CINT) | (cp[IDENT] == POINTER))
				sz = sz * INTWIDTH;
			fprintf(out, " =~data+%u\n", dp);
			dp = dp + sz;
		}
		cp = cp + SYMSIZ;
	}
	fprintf(out, "~eod =~data+%u\n", dp);
}

/*
 * output assembler comment
 */
comment(s)
char *s;
{
	fprintf(out, "; %s", s);
}

/*
 * output line of code
 */
outline(s)
char *s;
{
	fprintf(out, "\t%s\n", s);
}

/*
 * output line of code with integer value
 */
outval(s, n)
char *s;
int n;
{
	fprintf(out, s, n);
	lastop = ARITH;
}

/*
 * output line of code with string value
 */
outstr(s1, s2)
char *s1;
char *s2;
{
	fprintf(out, s1, s2);
	lastop = ARITH;
}

/*
 * pop secondary, and output line of JSR threaded code
 */
popcode(s)
char *s;
{
	pop();
	outcode(s);
}

/*
 * output line of JSR threaded code (arithmetic)
 */
outcode(s)
char *s;
{
	fprintf(out, "\tjsr %s\n", s);
	lastop = ARITH;
}

/*
 * output line of JSR threaded code (relational)
 */
relcode(s)
char *s;
{
	pop();
	fprintf(out, "\tjsr %s\n", s);
	lastop = REL;
}

#ifdef DEBUG
/*
 * output trace/debug call
 */
trace()
{
	outcode("__trace");
}
#endif

/*
 * output global label by symbol table entry
 */
label(name)
char *name;
{
	outstr("_%s\n", name);
}

/*
 * output compiler generated label
 */
outlab(label)
int label;
{
	outval("~%d\n", label);
}

/*
 * Load direct 8 or 16 bits into primary register
 */
getmem(lval)
int *lval;
{
	char *sym;

	sym = lval[SYMBOL];
	outstr("\tldx _%s\n", sym);
	if ((sym[IDENT] != POINTER) & (sym[TYPE] == CCHAR))
		outcode("ext");
	else
		outstr("\tldy _%s+1\n", sym);
}

/*
 * Given offset from SP, get address into primary register
 */
getloc(sym)
char *sym;
{
	int n;

	n = mgetw(sym + OFFSET);
	if (n & 0x8000)
		n = -((n ^ 0xFFFF) + 1);
	n = n - sp;
	if (n == 0)
		outcode("addr");
	else if (n < 256) {
		if ((n <= 12) & (n % 2 == 0))
			outval("\tjsr addr_%d\n", n / 2);
		else {
			outval("\tldx #%d\n", n);
			outcode("addr_b");
		}
	}
	else {
		immed(n);
		outcode("addr_w");
	}
}

/*
 * Store direct 8 or 16 bits from primary register
 */
putmem(lval)
int *lval;
{
	char *sym;

	sym = lval[SYMBOL];
	outstr("\tstx _%s\n", sym);
	if ((sym[IDENT] == POINTER) | (sym[TYPE] != CCHAR))
		outstr("\tsty _%s+1\n", sym);
}

/*
 * Store indirect 8 or 16 bits at address on top of stack
 */
putstk(typeobj)
char typeobj;
{
	if (typeobj == CCHAR)
		outcode("sind_b");
	else
		outcode("sind_w");
}

/*
 * Load indirect 8 or 16 bits at address in primary reg into primary register
 */
indirect(typeobj)
char typeobj;
{
	if (typeobj == CCHAR)
		outcode("lind_b");
	else
		outcode("lind_w");
}

/*
 * Call subroutine
 */
call(sname)
char *sname;
{
	outstr("\tjsr _%s\n", sname);
}

/*
 * Subroutine call to address on top of stack, return address left on stack
 */
callstk()
{
	popcode("scall");
}

/*
 * Jump to specified internal label number
 */
jump(label)
int label;
{
	outval("\tjmp ~%d\n", label);
}

sjump(s)
char *s;
{
	outstr("\tjmp _%s\n", s);
}

/*
 * Jump to specified label if primary reg is false (zero)
 */
testjump(label)
int label;
{
	outval("\tbne *+5\n\tjmp ~%d\n", label);
}

dotestjump(label)
int label;
{
	outval("\tbeq *+5\n\tjmp ~%d\n", label);
}

testnz(label)
int label;
{
	outval("\tjsr nz\n\tbne *+5\n\tjmp ~%d\n", label);
}

dotestnz(label)
int label;
{
	outval("\tjsr nz\n\tbeq *+5\n\tjmp ~%d\n", label);
}

/*
 * Modify stack pointer to new value indicated
 */
modstk(newsp)
int newsp;
{
	int k;

	k = newsp - sp;
	if (k == 0)
		return newsp;
	if (k < 0)
		decstk(k);
	else
		incstk(k);
	return newsp;
}

decstk(n)
int n;
{
	if ((n = (n & 0xFFFF ^ 0xFFFF) + 1) < 0xFF)
		outval("\tlda #%d\n\tjsr sdec_b\n", n);
	else {
		outval("\tlda #%d\n\tsta asave\n", n >> 8);
		outval("\tlda #%d\n\tjsr sdec_w\n", n & 0xFF);
	}
}

incstk(n)
int n;
{
	if (n < 0xFF) {
		if (n == 2)
			outcode("drop");
		else if (n == 4 | n == 6)
			outval("\tjsr drop%d\n", n >> 1);
		else
			outval("\tlda #%d\n\tjsr sinc_b\n", n);
	}
	else {
		outval("\tlda #%d\n\tsta asave\n", n >> 8);
		outval("\tlda #%d\n\tjsr sinc_w\n", n & 0xFF);
	}
}

/*
 * (dummy) start to main segment
 */
header()
{
	lastop = 0;
}

/*
 * swap primary and secondary
 */
swap()
{
	outcode("swap");
}

/*
 * load immediate by value
 */
immed(val)
int val;
{
	outval("\tldx #<%u\n", val);
	outval("\tldy #>%u\n", val);
}

/*
 * load immediate by name
 */
immediate(name)
char *name;
{
	outstr("\tldx #<_%s\n", name);
	outstr("\tldy #>_%s\n", name);
}

/*
 * load address of literal
 */
literal(lab, offset)
int lab;
int offset;
{
	outval("\tldx #<~%d+", lab);
	outval("%u\n", offset);
	outval("\tldy #>~%d+", lab);
	outval("%u\n", offset);
}

/*
 * push primary onto stack
 */
push()
{
	outcode("push");
	sp = sp - INTWIDTH;
}

/*
 * pop (compiler) stack
 */
pop()
{
	sp = sp + INTWIDTH;
}

/*
 * swap primary and top of stack
 */
swapstk()
{
	outcode("xchange");
}

/*
 * return from function
 */
ret()
{
	outline("rts");
}

/*
 * exit from program at arbitrary depth of function call nesting
 */
c_exit()
{
	outline("jmp _exit");
}

/*
 * scale primary by INTWIDTH
 */
scale(lval)
int *lval;
{
	char *p;

	if ((p = lval[SYMBOL]) != NULL) {
		if ((p[IDENT] == POINTER) | (p[IDENT] == ARRAY)) {
			if (p[TYPE] == CINT)
				outcode("scale2");
		}
	}
}

/*
 * add primary and top stack item
 */
add(lval)
int *lval;
{
	pop();
	scale(lval);
	outcode("add");
}

/*
 * subtract primary from top stack item
 */
sub(lval)
int *lval;
{
	pop();
	scale(lval);
	outcode("sub");
}

/*
 * multiply primary and top stack item
 */
mult()
{
	pop();
	outcode("mult");
}

/*
 * check 'usual' binary conversions
 */
uconv(lval1, lval2)
int *lval1, *lval2;
{
	char *p1, *p2;
	int t1, t2;

	if ((p1 = lval1[SYMBOL]) != NULL)
		t1 = p1[IDENT];
	else
		t1 = NULL;
	if ((p2 = lval2[SYMBOL]) != NULL)
		t2 = p2[IDENT];
	else
		t2 = NULL;
	return (t1 == POINTER) | (t2 == POINTER);
}

/*
 * divide top stack item by primary
 * --------------------------------
 * quotient in primary, rem in secondary
 */
div(lval1, lval2)
int *lval1, *lval2;
{
	if (uconv(lval1, lval2))
		popcode("udiv");
	else
		popcode("div");
}

/*
 * mod of top stack item divided by primary
 * ----------------------------------------
 * rem in primary, quotient in secondary
 */
mod(lval1, lval2)
int *lval1, *lval2;
{
	if (uconv(lval1, lval2))
		popcode("umod");
	else
		popcode("mod");
}

/*
 * inclusive or of primary and top stack item
 */
or()
{
	popcode("or");
}

/*
 * exclusive or of primary and top stack item
 */
xor()
{
	popcode("xor");
}

/*
 * logical and of primary and top stack item
 */
and()
{
	popcode("cand");
}

/*
 * arithmetic shift right of top stack item
 * ----------------------------------------
 * number of shifts in primary
 */
asr()
{
	popcode("casr");
}

/*
 * arithmetic shift left
 */
asl()
{
	popcode("casl");
}

/*
 * twos complement of primary
 */
neg()
{
	outcode("neg");
}

/*
 * increment primary by n
 */
inc(lval)
int *lval;
{
	char *p;

	p = lval[SYMBOL];
	if ((p[IDENT] == POINTER) & (p[TYPE] == CINT))
		outcode("inc2");
	else
		outcode("inc1");
}

/*
 * decrement primary
 */
dec(lval)
int *lval;
{
	char *p;

	p = lval[SYMBOL];
	if ((p[IDENT] == POINTER) & (p[TYPE] == CINT))
		outcode("dec2");
	else
		outcode("dec1");
}

/*
 * Conditional instructions
 * ------------------------
 * compare top stack item against primary, put 1 in primary if true, otherwise 0
 */

/*
 * == (equal)
 */
eq()
{
	relcode("eq");
}

/*
 * != (not equal)
 */
ne()
{
	relcode("ne");
}

/*
 * < (less than)
 */
lt(lval1, lval2)
int *lval1, *lval2;
{
	if (uconv(lval1, lval2))
		relcode("ult");
	else
		relcode("lt");
}

/*
 * <= (less than or equal)
 */
le(lval1, lval2)
int *lval1, *lval2;
{
	if (uconv(lval1, lval2))
		relcode("ule");
	else
		relcode("le");
}

/*
 * > (greater than)
 */
gt(lval1, lval2)
int *lval1, *lval2;
{
	if (uconv(lval1, lval2))
		relcode("ugt");
	else
		relcode("gt");
}

/*
 * >= (signed)
 */
ge(lval1, lval2)
int *lval1, *lval2;
{
	if (uconv(lval1, lval2))
		relcode("uge");
	else
		relcode("ge");
}

/*
 * literal definitions
 */
defbyte(c)
char c;
{
	outval("\t.byte %d\n", c);
}

defstorage(n)
int n;
{
	outval("\t*=*+%u\n", n);
}

/*
 * (dummy) end of assembly
 */
trailer()
{
	lastop = 0;
}
